/**
 * FractalBitmap.as
 * 
 * Copyright (c) 2008 Tom Beddard
 * http://www.subblue.com
 * 
 * Licensed under the MIT License: http://www.opensource.org/licenses/mit-license.php
 * 
 * Credits:
 * Bucket rendering implementation based on the implementation by Neave:
 *   http://www.neave.com/fractal/
 * 
 * References:
 *   http://en.wikipedia.org/wiki/Mandelbrot_set
 */

package com.subblue.fractal
{
	
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.Matrix;
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.PixelSnapping;
	import flash.display.Stage;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.utils.Timer;
	import flash.utils.getTimer;
	import com.subblue.utils.ColorUtil;
	import com.subblue.fractal.Mapping;
	import com.subblue.fractal.mappings.*;
	
	
	public class FractalBitmap extends Bitmap
	{
		// Class constants
		static public const BUCKET_SIZE:int = 64;
		static public const BUCKET_COMPLETE:String = "bucketComplete";
		static public const RENDER_QUEUE_STARTED:String = "renderQueueStarted";
		static public const RENDER_QUEUE_COMPLETE:String = "renderQueueComplete";
		static public const MAX_MAGNIFICATION:Number = 8e12;
		
		// Instance constants
		private   var DEFAULT_MAGNIFICATION = 1;
		private   var X_ORIGIN = -0.5;
		private   var Y_ORIGIN = 0;
		
		// Instance variables
		private   var _mapping:Mapping;
		protected var _maxIterations:int = 250;				// Maximum number of iterations
		protected var _bailout:Number = 128;				// Bailout value
		protected var _bailoutType:String = 'standard';		// Bailout type
		private   var _ox:Number = X_ORIGIN;				// x origin in imaginary plane
		private   var _oy:Number = Y_ORIGIN;				// y origin in imaginary plane
		private   var _magnification:Number = 1;			// magnification amount
		private   var _sampling:int = 1;					// Pixel sampling
		private   var _dx:Number;
		private   var _dy:Number;
		private   var _bucketSize:int = BUCKET_SIZE;
		
		private   var bucketQueue:Array = [];
		private   var rendering:Boolean = false;
		private   var bucketComplete:Boolean = false;
		private   var bucketCount:int = 0;
		
		private   var minReal:Number = 0.0;
		private   var minImaginary:Number = 0.0;
		private   var histogram:Array;				// Store the distribution of iterations
		
		private   var timer:Timer;
		private   var startTime:Number = 0.0;
		private   var _elapsedTime:Number = 0.0;
		
		
		
		/**
		 * Setup the fractal
		 * 
		 * @param	bitmapData		BitmapData to draw into
		 * @param	mapping			The mapping to use for this fractal
		 */
		public function FractalBitmap(bitmapData:BitmapData = null, mapping:Mapping = null)
		{
			super(bitmapData, PixelSnapping.ALWAYS, true);
			_mapping = mapping;
			initialiseTimer();
			addEventListener(BUCKET_COMPLETE, bucketCompleteListener);
		}
		
		
		/**
		 * Render the fractal
		 *	
		 * @param	region		Rectangle to render
		 */
		public function render(region:Rectangle = null):void
		{
			trace("Start render");
			if (!region) region = new Rectangle(0, 0, bitmapData.width, bitmapData.height);
			stopRenderQueue();
			prepareRenderQueue(region);
			startTimer();
			processRenderQueue();
		}
		
		
		/**
		 *	This is overridden by the subclasses that actually define the fractal formulas
		 */
		protected function formula(x:Number, y:Number):Object
		{
			return {};
		}
		
		
		/**
		 *	This is overridden by the subclasses that actually define the fractal formulas
		 */
		protected function bailoutCondition(aa:Number, bb:Number):Number
		{
			return (aa + bb);		
		}
		
		
		/**
		 *	Reset the fractal
		 */
		public function reset():void
		{
			stopRenderQueue();
			bitmapData.fillRect(new Rectangle(0, 0, bitmapData.width, bitmapData.height), 0xFF000000);
			ox = X_ORIGIN;
			oy = Y_ORIGIN;
			magnification = DEFAULT_MAGNIFICATION;
			render();
		}
		
		
		/**
		 *	Pan the bitmap object
		 */
		public function pan(dx:int, dy:int, dz:Number = 1, scrollBitmap:Boolean = true):void
		{
			if (dx != 0 || dy != 0)
			{
				if (scrollBitmap) bitmapData.scroll(-dx, -dy);
				_ox += (_dx / dz) * dx;
				_oy += (_dy / dz) * dy;
				
				if (rendering)
				{
					// We still haven't completed the previous render so need to redraw everything
					// so we don't get odd buckets left behind
					render();
					
				} else {
					// Render the uncovered areas
					var r1:Rectangle;
					var r2:Rectangle;
				
					if (dx != 0)
					{
						if (dx > 0)
						{
							r1 = new Rectangle(width - dx, 0, dx, height);
						} else {
							r1 = new Rectangle(0, 0, -dx, height);
						}
						prepareRenderQueue(r1);
						clear(r1);
					}

					if (dy != 0)
					{
						if (dy > 0)
						{
							r2 = new Rectangle(0, height - dy, width, dy);
						} else {
							r2 = new Rectangle(0, 0, width, -dy);
						}
					
						if (r1)
						{
							if (r1.x == 0) r2.x = r1.width;
							r2.width = width - r1.width;
							prepareRenderQueue(r2, false);
						} else {
							prepareRenderQueue(r2);
						}
					
						clear(r2);
					}
				}
				startTimer();
				processRenderQueue();
			}
		}
		
		
		/**
		 * Zoom the fractal
		 */
		public function zoom(dz:Number, mx:Number, my:Number):void
		{
			magnification *= dz;
			/*trace("Zoom: ", magnification);*/
			if (_magnification == MAX_MAGNIFICATION) return;
			
			bitmapData.lock();
			
			// Copy the bitmap, zoom and then merge back in
			var dx:int = bitmapData.width / 2 - mx;
			var dy:int = bitmapData.height / 2 - my;
			var zoomed:BitmapData = bitmapData.clone();
			var m:Matrix = new Matrix();
			m.scale(dz, dz);
			m.translate(mx - mx * dz, my - my * dz);
		
			if (dz < 1)
			{
				clear();
				dz = -(1 / (2 * dz));
			}
			
			bitmapData.draw(zoomed, m);
			zoomed.dispose();
			bitmapData.unlock();
			pan(-dx, -dy, dz, false);
			render();
		}
		
		
		/**
		 *	Resize the bitmap
		 */
		public function resize(w:int, h:int):void
		{
			bitmapData.lock();
			var original:BitmapData = bitmapData.clone();
			bitmapData = new BitmapData(w, h);
			bitmapData.draw(original);
			original.dispose();
			bitmapData.unlock();
		}
		
		
		/**
		 *	Return the real component value for the current pixel position
		 */
		public function real(value:int):Number
		{
			return minReal + (_dx * value);
		}
		
		
		/**
		 *	Return the imaginary component value for the current pixel position
		 */
		public function imaginary(value:int):Number
		{
			return minImaginary + (_dy * value);
		}
		
		
		/**
		 *	Get the previous render time
		 */
		public function renderTime():Number
		{
			return _elapsedTime;
		}
		
		
		
		////////////////////////////////////////////////////
		//
		// Private methods
		//
		////////////////////////////////////////////////////
		
		
		/**
		 *	Clear the bitmap
		 */
		private function clear(r:Rectangle = null):void
		{
			if (!r) r = new Rectangle(0, 0, bitmapData.width, bitmapData.height);
			bitmapData.fillRect(r, 0x00000000);
		}
		
		
		/**
		 *  Prepare the render bucket queue
		 */
		private function prepareRenderQueue(region:Rectangle, clear:Boolean = true):void
		{
			stage.focus = null;
			if (clear) bucketQueue = [];
			
			var bs:int = _bucketSize / _sampling;
			var xBuckets:int = Math.ceil(region.width / bs);
			var yBuckets:int = Math.ceil(region.height / bs);
			var l:int = region.left
			var t:int = region.top;
			var w:int, h:int;
			
			if (xBuckets > 1 && yBuckets > 1)
			{
				// Create the bucket list and reduce bucket sizes for edges if required.
				for (var i:int = 0; i < yBuckets; i++)
				{
					l = region.left;
					h = bs;
					if (t + bs > region.bottom) h = region.bottom - t;
				
					for (var j:int = 0; j < xBuckets; j++)
					{
						w = bs;
						if (l + bs > region.right) w = region.right - l;
						bucketQueue.push({x:l, y:t, w:w, h:h});
						l += w;
					}
					t += h;
				}
			} else {
				// Render region is too small for buckets
				bucketQueue.push({x:region.x, y:region.y, w:region.width, h:region.height});
			}
			
			rendering = true;
			bucketCount = 0;
		}
		
		
		/**
		 *  Process the render queue
		 */
		private function processRenderQueue():void
		{
			
			if (bucketCount < bucketQueue.length)
			{
				if (bucketCount == 0)
					dispatchEvent(new Event(RENDER_QUEUE_STARTED));
				var b:Object = bucketQueue[bucketCount];
				renderBucket( new Rectangle(b.x, b.y, b.w, b.h) );
				bucketCount++;
			} else {	
				stopRenderQueue();
			}
		}
		
		
		/**
		 *	Stop the rendering
		 */
		private function stopRenderQueue():void
		{
			stopTimer();
			bucketComplete = true;
			rendering = false;
			bucketCount = 0;
			dispatchEvent(new Event(RENDER_QUEUE_COMPLETE));
		}
		
		
		/**
		 *	Called when the fractal has finished rendering a bucket
		 */
		private function bucketCompleteListener(e:Event):void
		{
			bucketComplete = true;
		}
		
		
		/**
		 *  Listener to update the render queue
		 */
		private function renderQueueListener(e:TimerEvent):void
		{
			// Render the next bucket once the current one has completed
			if (bucketComplete && rendering)
			{
				bucketComplete = false;
				processRenderQueue();
				e.updateAfterEvent();
			}
		}
		
		
		/**
		 * Render a fractal bucket
		 */
		private function renderBucket(r:Rectangle = null):void
		{
			var i:int, j:int, n:int, ii:int, jj:int;
			var a:Number, b:Number, aa:Number, bb:Number, sdx:Number, sdy:Number;
			var samples:Array;
			var w:int = bitmapData.width;
			var h:int = bitmapData.height;
			var x1:Number = _ox - 2/_magnification;			// Left limit of x
			var x2:Number = _ox + 2/_magnification;			// Right limit of x
			var spanX:Number = x2 - x1;
			var spanY:Number = spanX * (h / w);
			var y1:Number = _oy - spanY / 2;
			var y2:Number = _oy + spanY / 2;
			var z:Object;
			
			// Offset x & y limits based on the region we want to render
			_dx = (spanX / w);								// x increment
			_dy = (spanY / h);								// y increment
			sdx = _dx / _sampling;							// x increment when super-sampling
			sdy = _dy / _sampling;							// y increment when super-sampling
			x1 += r.left * _dx;
			x2 -= (w - r.right) * _dx;
			y1 += r.top * _dy;
			y2 -= (h - r.bottom) * _dy;
			
			// Set min real & imaginary values for the current view so we can 
			// inspect any point later with real() and imaginary() methods
			if (r.left == 0 && r.top == 0)
			{
				minReal = x1;
				minImaginary = y1;
			}
			
			bitmapData.lock();
			bitmapData.fillRect(r, 0xFF000000);
			
			b = y1;
			for (j = r.top; j < r.bottom; j++)
			{
				a = x1;
				for (i = r.left; i < r.right; i++)
				{
					// Super sampling loop
					samples = [];
					aa = a;
					bb = b;
					
					for (jj = 0; jj < _sampling; jj++)
					{
						aa = a;
						for (ii = 0; ii < _sampling; ii++)
						{
							// Each pixel is coloured based on the time it takes the sequence 
		     				// to get to infinity. If we never get there then we are outside 
		     				// the set and colour the pixel black
			    			z = formula(aa, bb);

		     				if (z.n < _maxIterations)
							{
		     					// Outside mapping
								samples.push(_mapping.map(z.a, z.b, z.zz, z.n))
		     				} else {
								// Inside mapping - TODO
								samples.push(0xFF000000);
							}
		     				aa += sdx;
						}
						if (_sampling > 1) 
						{
							bb += sdy;
						}
					}
					
					if (_sampling > 1)
					{
						bitmapData.setPixel32(i, j, ColorUtil.average(samples));
					} else {
						bitmapData.setPixel32(i, j, samples[0]);
					}
					
					a += (_sampling * sdx);
				}
				b += (_sampling * sdy);
			}
			
			bitmapData.unlock();
			dispatchEvent(new Event(BUCKET_COMPLETE));
		}
		
		
		/**
		 *	Initialise the timer
		 */
		private function initialiseTimer():void
		{
			timer = new Timer(1, 0);
			timer.addEventListener(TimerEvent.TIMER, renderQueueListener);
		}
		
		
		/**
		 *	Start the event timer and the render timer
		 */
		private function startTimer():void
		{
			timer.reset();
			timer.start();
			startTime = getTimer();
		}
		
		
		/**
		 *	Return the elapsed time
		 */
		private function checkTimer():Number
		{
			return getTimer() - startTime;
		}
		
		
		/**
		 *	Stop the event timer and show render time
		 */
		private function stopTimer(msg:String = "Execution Time"):void
		{
			_elapsedTime = checkTimer();
			timer.stop();
			
			if (rendering)
			{
				if (_elapsedTime > 1000)
				{
					trace(msg, String(_elapsedTime / 1000) + "s");
				} else {
					trace(msg, String(_elapsedTime) + "ms");
				}	
			}
		}
		
		
		
		////////////////////////////////////////////////////
		//
		// Getter and Setter methods
		//
		////////////////////////////////////////////////////
		
		/**
		 *	Get and set maxIterations
		 */
		public function get maxIterations():int { 
			return _maxIterations; 
		}
		public function set maxIterations(value:int):void {
			if (value !== _maxIterations && _maxIterations >= 50) _maxIterations = value;
		}
		
		
		/**
		 *	Get and set bailoutValue
		 */
		public function get bailout():int { 
			return _bailout; 
		}
		public function set bailout(value:int):void {
			if (value !== _bailout) _bailout = value;
		}
		
		
		/**
		 *	Get and set bailoutType
		 */
		public function get bailoutType():String { 
			return _bailoutType; 
		}
		public function set bailoutType(value:String):void {
			if (value !== _bailoutType) _bailoutType = value;
		}
		
		
		/**
		 *	Get and set xOrigin
		 */
		public function get ox():Number { 
			return _ox; 
		}
		public function set ox(value:Number):void {
			if (value !== _ox) _ox = value;
		}
		
		
		/**
		 *	Get and set yOrigin
		 */
		public function get oy():Number { 
			return _oy; 
		}
		public function set oy(value:Number):void {
			if (value !== _oy) _oy = value;
		}
		
		
		/**
		 *	Get and set magnification amount
		 */
		public function get magnification():Number { 
			return _magnification; 
		}
		public function set magnification(value:Number):void {
			if (value !== _magnification) _magnification = value;
			if (_magnification > MAX_MAGNIFICATION * _sampling) _magnification = MAX_MAGNIFICATION / _sampling;
		}
		
		
		/**
		 *	Getters for the dx and dy increments
		 */
		public function get dx():Number { 
			return _dx; 
		}
		public function get dy():Number { 
			return _dy; 
		}
		
		
		/**
		 *	Get and set pixel sampling factor
		 */
		public function get sampling():Number { 
			return _sampling; 
		}
		public function set sampling(value:Number):void {
			if (value !== _sampling) _sampling = value;
		}
		
		
		/**
		 *	Get and set the mapping
		 */
		public function get mapping():Mapping { 
			return _mapping; 
		}
		public function set mapping(value:Mapping):void {
			if (value !== _mapping) _mapping = value;
		}
		
		
		/**
		 *	Get and set the bucket size
		 */
		public function get bucketSize():int { 
			return _bucketSize; 
		}
		public function set bucketSize(value:int):void {
			if (value !== _bucketSize) _bucketSize = value;
		}
		
		/**
		 *	Get the bitmap data object for image export
		 */
		public function get imageData():BitmapData{
			return bitmapData;
		}
	}
}