/**
 * Mapping.as
 * 
 * Copyright (c) 2008 Tom Beddard
 * http://www.subblue.com
 * 
 * Licensed under the MIT License: http://www.opensource.org/licenses/mit-license.php
 */

package com.subblue.fractal
{
	import com.subblue.fractal.ColorPalette;
	import com.subblue.utils.ColorUtil;

    public class Mapping
    {
        public static const LINEAR:int 		= 0;
        public static const SQR:int 		= 1;
        public static const SQR_ROOT:int 	= 3;
        public static const CUBE:int 		= 2;
        public static const CUBE_ROOT:int 	= 4;
        public static const LOG:int 		= 5;
        public static const EXP:int 		= 6;
		
        private   var _transfer:int = LINEAR;
		protected var _palette:Array;
		protected var _size:int = 200;
		private	  var _cycle:Boolean = true;
		private	  var _mapping:Array = [];
		
		
		/**
		 *	Create an array of colours for the colour mapping
		 */
		public function Mapping(palette:Array = null, size:int = 0)
        {
            if (size > 0) _size = size;
			if (!palette) this.palette = ColorPalette.DEFAULT;
        }
		
		
		/**
		 *	Get the colour for the requested index. 
		 *	Will loop if the index is greater than the mapping size.
		 */
		public function color(i:int):uint
		{
			var idx:Number = Math.round(transfer(i));
			
			if (_transfer > 2)
			{
				// Float index
				var c1 = _mapping[int(idx) % _size];
				var c2 = _mapping[(int(idx) + 1) % _size];
				return ColorUtil.blend(c1, c2, idx - int(idx));
			} else {
				// Integer index
				return _mapping[int(idx) % _size];
			}
		}
		
		
		/**
		 *	Return the colour value for the required index. No longer used?
		 *	This is the linear transfer
		 *	
		 *	@param		r		Real component
		 *	@param		i		Imaginary component
		 *	@param		n		Number of iterations taken
		 */
		public function map(r:Number, i:Number, zz:Number, n:int):uint
		{
			return _mapping[n % _size];
		}
		
		
		
		/**
		 *	Generate the colour map array
		 */
		private function generateMap():void
		{
			// Steps between each colour band. Blend the first and last colours if cycle is set
			var steps:int = _size / (_palette.length + (_cycle ? 1 : -1));
			var i:int;
			_mapping = [];
			
			for (i = 1; i < _palette.length; i++)
			{
				_mapping = _mapping.concat(ColorUtil.blendArray(_palette[i-1], _palette[i], (steps - 1), false));
			}
			
			if (_cycle)
			{
				// Blend from the last colour to the first
				steps = _size - _mapping.length;
				_mapping = _mapping.concat(ColorUtil.blendArray(_palette[i-1], _palette[0], (steps - 1), false));	
			}
			/*trace("Mapping sizes:", _size, _mapping.length);*/
		}
		
		
		
		/**
		 *	Get and set the palette array of colours
		 */
		public function get palette():Array {
			return _palette;
		}
		public function set palette(value:Array):void {
			_palette = value;
			generateMap();
		}
		
		/**
		 *	Get and set the size of the mapping array
		 */
		public function get size():int {
			return _size;
		}
		public function set size(value:int):void {
			if (value !== _size)
			{
				_size = value;
				generateMap();
			}
		}
		
		/**
		 *	Get and set colour cycling option
		 */
		public function get cycle():Boolean { 
			return _cycle; 
		}
		public function set cycle(value:Boolean):void {
			if (value !== _cycle)
			{
				_cycle = value;
				generateMap();
			}
		}
		
		/**
		 *	Get and set transfer function
		 */
		public function get transferFunction():Number { 
			return _transfer; 
		}
		public function set transferFunction(value:Number):void {
			trace("Transfer function: ", value);
			if (value !== _transfer) _transfer = value;
		}
		
		
		
		/**
		 *	Decide how the incoming request for a colour should be mapped onto the colour array
		 */
		protected function transfer(value:int):Number
		{
			switch(_transfer)
			{
				case LINEAR:
					return value;
				case SQR:
					return value * value;
				case SQR_ROOT:
					return Math.sqrt(value);
				case CUBE:
					return Math.pow(value, 3);
				case CUBE:
					return Math.pow(value, 1/3);
				case EXP:
					return Math.exp(value);
				case LOG:
					return Math.log(value);
				default:
					return value;
			}
		}
		
		/**
		 * Returns a representation of the object as a string
		 */
		public function toString():String {
			return "Mapping";
		}
		
    }
}
